/* QuakeMud, a silly mud using libhacknslash
 * Copyright (C) 2000 David Wynn (aka kismet), of the OSOAL collective
 * send abuse to:    david@osoal.org.nz
 * http requests to: web.osoal.org.nz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <glib.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <dirent.h>

#include <hacknslash/server.h>
#include <hacknslash/misc.h>
#include <hacknslash/colors.h>

#include "players.h"
#include "rooms.h"
#include "weapons.h"
#include "commands.h"
#include "game.h"

/* default config parameters */
gint     game_frag_limit         = 50;
gint     game_time_limit         = 15 * 60;
gint     game_port               = 4950;
gint     game_max_clients        = 0;
gboolean game_weapon_persistance = FALSE;
gboolean game_powerups           = TRUE;
gchar   *game_rpassword          = NULL;

/* INSTALLDIR defined in Makefile */
#define DEFAULT_MAPDIR INSTALLDIR "/maps"

gchar   *game_mapdir             = NULL;

#define DEFAULT_CONFIG INSTALLDIR "/config"

gchar   *game_config             = NULL;

#define DEFAULT_LOG INSTALLDIR "/log"

gchar   *game_log_file = NULL;
static FILE *log_file = NULL;

#define WELCOME_MESSAGE INSTALLDIR "/welcome"



#define NAME_MAX_CHARS 16

/* make sure the name a player entered is valid
 * (letters and numbers and spaces only) 
 */
static gboolean
_name_is_valid (gchar *name)
{
  gint count = 0;
  while (*name && count < NAME_MAX_CHARS)
    {
      if (!isprint (*name))
	return FALSE;
      name++;
      count++;
    }
  return TRUE;
}



/* login sequence */
static void _get_color  ( Player_t * p );
static void
_input_color (Player_t *p, gchar *line)
{
  gchar ch = tolower (line[0]);
  if (ch == 'y')
    {
      ((Sock_t *) p)->colors = TRUE;
    }
  else if (ch != 'n')
    {
      _get_color (p);
      return;
    } 

  /* end of login sequence, so announce player's joining */
  game_broadcastf (COL_EVENT, "\n`%s' has joined the game.", p->name);
  p->handler = NULL;
  player_to_game (p);
  player_spawn (p);
  players_send_prompts ();
}

static void
_get_color (Player_t *p)
{
  player_send (p, -1, "Would you like ansi color?");
  p->handler = _input_color;
}

/* */
static void _get_name   ( Player_t * p );
static void
_input_name (Player_t *p, gchar *line)
{
  if (line[0] == '\0')
    {
      server_disconnect ((Sock_t *) p);
      return;
    }
  
  if (!_name_is_valid (line))
    {
      player_send( p, -1, "That name is invalid." );
    }
  else if (player_find (line, FALSE))
    {
      player_send (p, -1, "That name is taken, please enter another.");
    }
  else
    {
      player_set_name (p, line);
      player_set_team (p, line); /* teams are "implemented" ie, 1 player per team */
      _get_color (p);
      return;
    }
}

static void
_get_name (Player_t *p)
{
  player_send (p, -1, "Enter your nickname (up to 16 chars)");
  p->handler = _input_name;
}



/* server event handlers */
static void
connect_handler (Player_t *p, gpointer data)
{
  player_spool (p, WELCOME_MESSAGE);
  _get_name (p);
  player_send_prompt (p);
}

static void
disconnect_handler (Player_t *p, gpointer data)
{
  if (!p)
    return;
  if (p->state != PLAYER_STATE_LOGIN)
    {
      game_broadcastf (COL_EVENT, "\n`%s' has disconnected.", p->name);
      player_to_limbo (p);
    }
  player_from_game (p);
  players_send_prompts ();
}



CmdList_t game_cmd_list; /* in-game commands (while alive) */
CmdList_t limbo_cmd_list; /* commands while dead */

/* parse a text line from a player */
static void
read_handler (Player_t *p, gpointer data)
{
  gchar *line;
  p->send_prompt = TRUE;

  /* remember, Player is derived from Sock */
  line = ((Sock_t *) p)->line;
  g_strstrip (line);

  if (p->handler)
    p->handler (p, line);
  else
    {
      if (p->state == PLAYER_STATE_LIMBO)
	{
	  if (line[0] == '\0')
	    player_spawn (p);
	  else
	    cmd_exec (&limbo_cmd_list, (Object_t *) p, line);
	}
      else if (p->state == PLAYER_STATE_ALIVE)
	{
	  cmd_exec (&game_cmd_list, (Object_t *) p, line);
	}
    }

  player_send_prompt (p);
}



void
game_log (gchar *str)
{
  fprintf (log_file, "%s\n", str);
  fflush (log_file);
}

void
game_log_init (void)
{
  log_file = fopen (game_log_file, "wt");
  fprintf (stdout, "logging to %s\n", game_log_file);
  if (!log_file)
    log_file = stdout;
}

/* send `text' to all players and to stdout */
void
game_broadcast (gint col, gchar *text)
{
  players_send (col, text);
  while (*text == '\n')
    text++;
  game_log (text);
}

void
game_broadcastf (gint col, gchar *format,...)
{
  gchar *text;
  va_list args;
  va_start (args, format);
  text = g_strdup_vprintf (format, args);
  va_end (args);
  game_broadcast (col, text);
  g_free (text);
}



/* */
static GSList * _maps = NULL;        /* list of (char *), these are the maps in rotation  */
static GSList * _map_current = NULL; /* the link of the current map */
static time_t   _map_start = 0;      /* the time when we started the current map */

void
game_maplist_free ()
{
  GSList *scan;
  for (scan = _maps; scan; scan = scan->next)
    g_free (scan->data);
  g_slist_free (_maps);
  _maps = _map_current = NULL;
}

void
game_maplist_load ()
{
  Block_t *block;
  GSList *scan;

  game_maplist_free ();
    
  block = block_file_get ("MapList", NULL, game_config);  
  if (!block)
    {
      game_log ("Err0r! couldn't find block `MapList' in config file\n");
      exit (EXIT_FAILURE);
    }

  for (scan = block->vars; scan; scan = scan->next)
    {
      BlockVar_t *var = scan->data;
      if (!strcasecmp (var->label, "map"))
	_maps = g_slist_prepend (_maps, g_strdup (var->value));
    }

  block_free (block);
}



/* */
void
game_config_load ()
{
  Block_t *block;
  GSList *scan;

  block = block_file_get ("Config", NULL, game_config);  
  if (block)
    {
      for (scan = block->vars; scan; scan = scan->next)
	{
	  BlockVar_t *var = scan->data;
	  if (!strcasecmp (var->label, "frag_limit"))
	    game_frag_limit = atoi( var->value );
	  else if (!strcasecmp (var->label, "time_limit"))
	    game_time_limit = atoi( var->value ) * 60;
	  else if (!strcasecmp (var->label, "port"))
	    game_port = atoi (var->value);
	  else if (!strcasecmp (var->label, "max_clients"))
	    game_max_clients = atoi (var->value);
	  else if (!strcasecmp (var->label, "weapon_persistance"))
	    misc_strtobool (&game_weapon_persistance, var->value);
	  else if (!strcasecmp (var->label, "powerups"))
	    misc_strtobool (&game_powerups, var->value);
	  else if (!strcasecmp (var->label, "log"))
	    {
	      g_free (game_log_file);
	      game_log_file = g_strdup (var->value);
	    }
	  else if (!strcasecmp (var->label, "rpassword"))
	    {
	      g_free (game_rpassword);
	      game_rpassword = g_strdup (var->value);
	    }
	}
      
      block_free (block);  
    }
  else
    game_log ("Warning! couldn't find block `Config' in config file\n");

  game_maplist_load ();
}

static void 
game_map_stop ()
{
  players_to_limbo ();
  room_db_free ();
}

Block_t *
game_map_info (gchar *map)
{
  gchar dir[64];
  Block_t *block;
  snprintf (dir, 64, "%s/%s/info", game_mapdir, map);
  block = block_file_get ("MapInfo", NULL, dir);  
  return block;
}

/* */
static int
_filter (struct dirent *ent)
{
  return (ent->d_name[0] != '.');
}

void
game_maps_list (Player_t *p)
{
  struct dirent ** ents;
  int n = scandir (game_mapdir, &ents, _filter, alphasort);
  if (n > 0)
    {
      int i = 0;
      player_send (p, COL_LTGREY, "\nMap list\n");
      while (i < n)
	{
	  Block_t *block = game_map_info (ents[i]->d_name);
	  if (block)
	    {
	      BlockVar_t *var;
	      var = block_get_var (block, "name");
	      if (var)
		player_sendf (p, -1, "  \"%s\" (%s)\n", ents[i]->d_name, var->value);
	      block_free (block);
	    }
	  i++;
	}
    }
}

void
game_map_change (gchar *map)
{
  gchar dir[64];
  Block_t *block;

  if (NULL != (block = game_map_info (map)))
    {
      BlockVar_t *var;

      var = block_get_var (block, "name");
      if (var)
	{
	  game_broadcastf (COL_LTGREY, "\nChanging maps...\n\n%s (%s)\n", map, var->value);
	  game_map_stop ();

	  snprintf (dir, 64, "%s/%s", game_mapdir, map);
	  room_db_load (dir);
	}      

      block_free (block);
    }
}

void 
game_map_rotate ()
{
  if (_map_current)
    _map_current = _map_current->next;
  if (!_map_current)
    _map_current = _maps;
  if (!_map_current)
    {
      game_log ("Error! No maps!\n");
      exit (EXIT_FAILURE);
    }

  game_map_change (_map_current->data);

  _map_start = time (NULL);
}

static gboolean _game_reload_config = FALSE;

/* */
gboolean
timeout_func (gpointer data)
{
  Player_t *p;
  gboolean new_map = FALSE;
  time_t t = time (NULL);

  /* if we got a sighup, reload the config (map list and other stuff) */
  if (_game_reload_config)
    {
      _game_reload_config = FALSE;
      game_broadcast (COL_LTGREY, "\nSighup received, reloading config...");
      game_config_load ();
      game_map_rotate ();
      //return TRUE;
    }

  /* tick */
  players_think ();
  rooms_think ();
  players_send_prompts ();

  /* check for a map rotation */
  p = player_winning ();
  if (p)
    {
      if (game_frag_limit > 0)
	{
	  if (p->frags >= game_frag_limit)
	    {
	      game_broadcast (COL_LTGREY, "\nFrag limit hit!");
	      new_map = TRUE;
	    }
	}
      if (game_time_limit && !new_map)
	{
	  if (t - _map_start >= game_time_limit)
	    {
	      game_broadcast (COL_LTGREY, "\nTime limit hit!");
	      new_map = TRUE;
	    }
	}

      if (new_map)
	{
	  game_broadcastf (COL_LTGREY, "\n`%s' won the round with %d frags!", p->name, p->frags);
	  game_map_rotate ();
	  new_map = FALSE;
	}
    }

  return TRUE;
}



/* */
static void (*_sighup_old)( gint );

static void
sighup_handler (gint sig)
{
  _game_reload_config = TRUE;
}



/* */
int
main (void) {
  GMainLoop * loop;

  /* <sailor moon> Okay everyone, it's time for daemon mode! 
   * <chorus> RIGHT!
   * <sailor venus> You have used your tty for EvIL!
   */
  if (fork () != 0)
    exit (0);
  setsid ();
  close (STDIN_FILENO);
  signal (SIGHUP, SIG_IGN);
  if (fork () != 0)
    exit (0);

  /* reload config files on sighup */  
  _sighup_old = signal (SIGHUP, sighup_handler);
  
  srandom (time (NULL)); /* meh */

  /* config defaults */
  game_config = g_strdup (DEFAULT_CONFIG);
  game_mapdir = g_strdup (DEFAULT_MAPDIR);
  game_log_file = g_strdup (DEFAULT_LOG);

  game_config_load ();

  /* open the log file */
  game_log_init ();

  /* */
  game_log ("QuakeMUD v" VERSION "\n");

  /* hook server events */
  server_init ((SockClass_t *) player_class ());
  server_set_connect_func ((SockFunc_t) connect_handler, NULL);
  server_set_read_func ((SockFunc_t) read_handler, NULL);
  server_set_disconnect_func ((SockFunc_t) disconnect_handler, NULL);
  server_open (game_port, game_max_clients);

  /* add think timeout */
  g_timeout_add (1000, timeout_func, NULL);

  /* initialize the room database */
  room_db_init ();

  /* initialize command lists */  
  cmd_list_init (&limbo_cmd_list, sizeof (QCmd_t));
  cmd_list_add (&limbo_cmd_list, (Cmd_t *) limbo_commands);

  cmd_list_init (&game_cmd_list, sizeof (QCmd_t));
  cmd_list_add (&game_cmd_list, (Cmd_t *) game_commands);
  cmd_list_add (&game_cmd_list, (Cmd_t *) limbo_commands);

  /* load first map in maplist */
  game_map_rotate ();

  loop = g_main_new (TRUE);
  g_main_run (loop);

  return 0;
}
