/* QuakeMud, a silly mud using libhacknslash
 * Copyright (C) 2000 David Wynn (aka kismet), of the OSOAL collective
 * send abuse to:    david@osoal.org.nz
 * http requests to: web.osoal.org.nz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>
#include <time.h>

#include <hacknslash/array.h>
#include <hacknslash/misc.h>
#include <hacknslash/colors.h>

#include "players.h"
#include "weapons.h"
#include "game.h"

/* */
#define HIGHSCORE_DISPLAY 10

const gchar *powerup_name[ NUM_POWERUPS ] =
{
  "quad", "pent", "eyes", "regen", "haste", "suit"
};

static Player_t *_winner = NULL;
static Player_t *_zero = NULL;

static void
_send_scores (Player_t *p, gint max)
{
  Player_t *scan;
  gint i;
  time_t t = time (NULL);
  for (i = 0, scan = _winner;
       (max < 0 || i < max) && scan;
       i++, scan = scan->down)
    {
      player_sendf (p, -1, " Frags:%3d Time:%3u `%s'\n", 
		    scan->frags, 
		    (guint) (t - scan->join_time) / 60,
		    scan->name);
    }
}

void
player_send_players (Player_t *p)
{
  _send_scores (p, -1);
}

void
player_send_scores (Player_t *p)
{
  player_sendf (p, COL_LTGREY, "\nTop %d Rankings\n", HIGHSCORE_DISPLAY);
  _send_scores (p, HIGHSCORE_DISPLAY);
}

Player_t *
player_winning ()
{
  return _winner;
}

static void
_highscore_move_zero ()
{
  if (_zero)
    {
      if (_zero->frags == 0)
	return;
      if (_zero->frags < 0)
	{
	  while (_zero->up && _zero->up->frags < 0)
	    _zero = _zero->up;
	} 
      else
	{
	  while (_zero->down && _zero->down->frags >= 0)
	    _zero = _zero->down;
	}
    }
}

/* moves `p' up the high scores if necessary */
static void
_highscore_up (Player_t *p)
{
  Player_t *up;
  while (1)
    {
      up = p->up;
      if (up && p->frags > up->frags)
	{
	  /* swap positions with player above */
	  p->up = up->up;
	  up->down = p->down;
	  if (up->up) 
	    up->up->down = p;
	  else
	    _winner = p;
	  if (p->down) 
	    p->down->up = up;
	  p->down = up;
	  up->up = p;
	} 
      else
	break;
    }
  _highscore_move_zero ();
}

/* moves `p' down the high scores if necessary */
static void
_highscore_down (Player_t *p)
{
  Player_t *down;
  while (1)
    {
      down = p->down;
      if (down && p->frags < down->frags)
	{
	  /* swap positions with player below */
	  p->down = down->down;
	  down->up = p->up;
	  if (down->down) 
	    down->down->up = p;
	  if (p->up) 
	    p->up->down = down;
	  else
	    _winner = down;
	  p->up = down;
	  down->down = p;
	}
      else
	break;
    }
  _highscore_move_zero ();
}



/* */
void
player_set_defaults (Player_t *p)
{
  gint i;
  p->attacking = NULL;
  p->attackers = NULL;
  p->move = MOVE_MAX;
  p->health = 100;
  p->armor = 0;
  p->air = MAX_AIR;
  p->weapon = WEAPON_SHOTGUN;
  p->weapons = (1 << WEAPON_AXE) | (1 << WEAPON_SHOTGUN);
  p->ammo[ AMMO_SHELLS ] = 50;
  p->ammo[ AMMO_NAILS ] = 0;
  p->ammo[ AMMO_ROCKETS ] = 0;
  p->ammo[ AMMO_SLUGS ] = 0;
  p->ammo[ AMMO_CELLS ] = 0;
  p->tell_me_when_i_can_move = FALSE;
  for (i = 0; i < NUM_POWERUPS; i++)
    p->powerup[i] = 0;
  p->keys = 0;
}

/* */
static SockClass_t * _parent_class = NULL;
static PlayerClass_t _pclass;

static void
_player_init (Player_t *p)
{
  if (((ObjectClass_t *) _parent_class)->init)
    ((ObjectClass_t *) _parent_class)->init ((Object_t *) p);
  p->join_time = 0;
  p->state = PLAYER_STATE_LOGIN;
  p->handler = NULL;
  p->name = NULL;
  p->team = NULL;
  p->frags = 0;
  p->send_prompt = FALSE;
  p->admin = FALSE;
  p->peace_mode = FALSE;
  p->tell_me_when_i_can_talk = FALSE;
  p->room = NULL;
  p->up = p->down = NULL;
  p->say = SAY_MAX;
  player_set_defaults (p);
}

static void
_player_destroy (Player_t *p)
{
  player_from_game (p);
  g_free (p->name);
  if (((ObjectClass_t *) _parent_class)->destroy)
    ((ObjectClass_t *) _parent_class)->destroy ((Object_t *) p);
}

PlayerClass_t *
player_class ()
{
  if (!_parent_class)
    {
      ObjectClass_t *oclass = (ObjectClass_t *) &_pclass;
      oclass->name = "Player";
      oclass->size = sizeof (Player_t);
      oclass->init = (ObjectFunc_t) _player_init;
      oclass->destroy = (ObjectFunc_t) _player_destroy;
      oclass->parent_class = (ObjectClass_t *) _parent_class = sock_class ();
    }
  return &_pclass;
};

void
player_set_name (Player_t *p, gchar *name)
{
  g_free (p->name);
  p->name = g_strdup (name);
}

void
player_set_team (Player_t *p, gchar *team)
{
  g_free (p->team);
  p->team = g_strdup (team);
}

Player_t *
player_get (guint id)
{
  return (Player_t *) server_get_sock (id);
}

static void
_highscores_add (Player_t *p)
{
  if (!_zero)
    _zero = _winner = p;
  else
    {
      if (_zero->frags >= 0)
	{
	  if (_zero->down)
	    _zero->down->up = p;
	  p->down = _zero->down;
	  _zero->down = p;
	  p->up = _zero;
	  _zero = p;
	  
	}
      else
	{
	  if (_zero->up)
	    _zero->up->down = p;
	  else
	    _winner = p;
	  p->up = _zero->up;
	  _zero->up = p;
	  p->down = _zero;
	  _zero = p;
	}
    }
}

static void
_highscores_remove (Player_t *p)
{
  /* remove player from score ranking */
  if (p->down)
    p->down->up = p->up;
  if (p->up)
    p->up->down = p->down;
  else
    _winner = p->down;
  if (_zero == p)
    {
      if (p->up)
	_zero = p->up;
      else
	_zero = p->down;
    }
}

void
player_spool (Player_t *p, gchar *file)
{
  FILE *f;
  gchar buf[512];
  gint i;
  f = fopen (file, "r");
  if (f)
    {
      player_send (p, -1, "\n");
      while (1)
	{
	  i = fread (buf, 1, 511, f);
	  if (i <= 0)
	    break;
	  buf[i] = 0;
	  player_send (p, -1, buf);
	}
      fclose (f);
    }
}

void
player_to_game (Player_t *p)
{
  if (p->state != PLAYER_STATE_LIMBO &&
      p->state != PLAYER_STATE_ALIVE)
    {
      _highscores_add (p);
      p->state = PLAYER_STATE_LIMBO;
    }
}

void
player_from_game (Player_t *p)
{
  if (p->state != PLAYER_STATE_LOGIN &&
      p->state != PLAYER_STATE_REMOVED)
    {
      player_to_limbo (p);
      _highscores_remove (p);
      p->state = PLAYER_STATE_REMOVED;
    }
}

void
player_to_limbo (Player_t *p)
{
  GSList *scan;
  gint i;
  if (!p || !(p->room) || p->state != PLAYER_STATE_ALIVE)
    return;
  room_player_remove (p->room, p);
  player_set_attack (p, NULL);
  for (scan = p->attackers; scan; scan = scan->next)
    {
      Player_t *p = scan->data;
      p->attacking = NULL;
      player_set_attack (p, NULL);
    }
  g_slist_free (p->attackers);
  p->state = PLAYER_STATE_LIMBO;
  for (i = 0; i < NUM_POWERUPS; i++)
    p->powerup[i] = 0;
}

/* */
static void
_pull_player (Player_t *p, gpointer data)
{
  player_to_limbo (p);
  p->frags = 0;
  p->join_time = 0;
}

void 
players_to_limbo ()
{
  server_foreach ((SockFunc_t) _pull_player, NULL);
}

/* set `player' to attack `target' */
void
player_set_attack (Player_t *p, Player_t *target)
{
  if (p->attacking)
    {
      p->attacking->attackers = g_slist_remove (p->attacking->attackers, p);
    }
  p->attacking = target;
  if (target == p)
    target = NULL;
  if (target)
    {
      target->attackers = g_slist_prepend (target->attackers, p);
      if (target->attacking == NULL)
	{
	  target->attacking = p;
	  p->attackers = g_slist_prepend (p->attackers, target);
	}
    }
}

#define player_hit( PLAYER, TARGET, AIM )                         \
   ((random()%100) <                                              \
    CLAMP( (AIM) - ((TARGET)->powerup[ POWERUP_EYES ] ? 15 : 0 ), \
	   10,                                                    \
	   80 ))

static gchar *hit_severity[ 5 ] =
{
  "\nYou barely hit `%s'.",
  "\nYou hit `%s'.",
  "\nYou hit `%s'.",
  "\nYou hit `%s' good.", 
  "\nYou hit `%s' right between the eyes."
};

void
player_died (Player_t *p)
{
  player_to_limbo (p);
  player_send (p, COL_BAD, "\n\nYou have died.\nPress enter or enter `spawn' to respawn.\n");
  player_send_scores (p);
}

Item_t *
player_backpack (Player_t *p)
{
  Item_t *item = item_alloc ();
  gint i;
  item->name = g_strdup ("backpack");
  item->weapons = p->weapons;
  item->armor = 0;
  item->health = 0;
  item->keys = p->keys;
  item->spawn_interval = 0;
  item->spawn_time = 0;
  item->die_time = time (NULL) + 180;
  item->always_pickup = TRUE;
  for (i = 0; i < NUM_AMMO_TYPES; i++)
    item->ammo[i] = p->ammo[i];
  return item;
}

static gint
_player_damage_armor (Player_t *p, gint damage)
{
  if (p->armor)
    {
      gint absorb, d = damage - damage / 3;
      absorb = MIN(d, p->armor);
      p->armor -= absorb;
      damage -= absorb;
    }
  return damage;
}

void
player_take_damage (Player_t *p, Player_t *source, gint damage)
{
  gint dam_mod = random () % 5;
  
  /* calculate the real damage */
  damage = ((gfloat) ((dam_mod+5)*10+(random () % 10)+1) / 100.0) * damage;

  if (p->state != PLAYER_STATE_ALIVE)
    return;

  if (p->powerup[ POWERUP_PENT ])
    {
      player_sendf (p, COL_GOOD, "\nYour pentagram protects you from damage\n");
      return;
    }

  if (p == source)
    {
      player_sendf (p, COL_BAD,
		    "\nYou have been hit by your own %s for %d damage",
		    weapon_name[ p->weapon ],
		    damage);
    }
  else
    {
      if (p->powerup[ POWERUP_EYES ])
	player_send (source, COL_GOOD, "\nYou hit the pair of eyes.");
      else
	player_sendf (source, COL_GOOD, hit_severity[ dam_mod ], p->name);
      if (source->powerup[ POWERUP_EYES ])
	player_sendf (p, COL_BAD,
		      "\nYou have been hit for %d damage from someone's %s.",
		      damage,
		      weapon_name[ source->weapon ]);
      else
	player_sendf (p, COL_BAD,
		      "\nYou have been hit by `%s's %s for %d damage.",
		      source->name,
		      weapon_name[ source->weapon ],
		      damage);
    }

  p->health -= _player_damage_armor (p, damage);

  if (p->health <= 0)
    {
      if (source != p)
	{
	  game_broadcastf (COL_EVENT,
			   obituary[ source->weapon ][ 0 ], 
			   p->name,
			   source->name);

	  player_give_ammo (source, p->ammo);
	  player_give_weapons (source, p->weapons);
	  player_give_keys (source, p->keys);
	  source->frags++;
	  _highscore_up (source);
	}
      else
	{
	  game_broadcastf (COL_EVENT, obituary[ p->weapon ][ 1 ], p->name);
	  p->frags--;
	  _highscore_down (p);
	  /* drop the player's backpack in the room */
	  room_item_add (p->room, player_backpack (p));
	}

      player_died (p);
    }
}

void
player_suicide (Player_t *p)
{
  if (!p || p->state != PLAYER_STATE_ALIVE)
    return;
  game_broadcastf (COL_EVENT, "\n`%s' tastes gun.", p->name);
  p->frags--;
  _highscore_down (p);
  /* drop the player's backpack in the room */
  room_item_add (p->room, player_backpack (p));
  player_died (p);
}

void
player_choose_weapon (Player_t *p)
{
  gint i;
  for (i = WEAPON_BFG; i >= 0; i--)
    {
      if (player_has_weapon (p, i))
	{
	  if (weapons[i].ammo < 0 || p->ammo[ weapons[i].ammo ] > weapons[i].ammo_use)
	    {
	      p->weapon = i;
	      return;
	    }
	}
    }
  p->weapon = 0;
}

/* make `player' attack it's target(s) */
void
player_attack (Player_t *p, Player_t *target)
{
  GSList *targets, *scan;
  Weapon_t *weapon;
  Room_t *room;
  gboolean splash_hit_self = FALSE;

  if (!p || !target || p->state != PLAYER_STATE_ALIVE)
    return;

  weapon = &(weapons[ p->weapon ]);
  room = p->room;

  if (!room || target->room != room)
    return;

  /* there must always be enough ammo before player_attack() is called */
  if (weapon->ammo >= 0 && weapon->ammo_use)
    {
      p->ammo[ weapon->ammo ] -= weapon->ammo_use;
    }

  /* do splash damage */
  if (weapon->area > 1 && weapon->splash > 0)
    {
      /* get list of players deal splash damage to
       * this is icky and should be rewritten (among other stuff)
       */
      if (weapon->can_hit_self == FALSE)
	room_player_remove (room, p);
      room_player_remove (room, target);
      targets = misc_slist_choose (room->players, (weapon->area) - 1);
      room_player_add (room, target);
      if (weapon->can_hit_self == FALSE)
	room_player_add (room, p);
      
      /* */
      for (scan = targets; scan; scan = scan->next)
	{
	  Player_t *t = scan->data;
	  if (t)
	    {
	      if (player_hit (p, t, weapon->splash_aim))
		{
		  if (t == p)
		    splash_hit_self = TRUE;
		  else
		    player_take_damage (t, p, weapon->splash * (p->powerup[ POWERUP_QUAD ] ? 4 : 1) );
		}
	    }
	}
      g_slist_free (targets);
    }
  
  /* attack the immediate target */
  if (player_hit (p, target, weapon->aim))
    player_take_damage (target, p, weapon->damage * (p->powerup[ POWERUP_QUAD ] ? 4 : 1));
  else
    {
      if (target->powerup[ POWERUP_EYES ])
	player_send (p, COL_EVENT, "\nYou miss the pair of eyes.");
      else
	player_sendf (p, COL_EVENT, "\nYou miss `%s'.", target->name);
      
      if (p->powerup[ POWERUP_EYES ])
	player_send (target, COL_EVENT, "\nThe pair of eyes missed you.");
      else
	player_sendf (target, COL_EVENT, "\n`%s' missed you.", p->name);
    }
  
  /* attack self with splash (afterwards) 
   * its a nasty hack to handle the case when the player kills himself
   * with splash damage before he can hit the actual target.
   */
  if (splash_hit_self)
    player_take_damage (p, p, weapon->splash*(p->powerup[ POWERUP_QUAD ] ? 4 : 1));

  if (p->state != PLAYER_STATE_ALIVE)
    return;

  /* use up ammo */
  if (weapon->ammo >= 0)
    {
      if (p->ammo[ weapon->ammo ] < weapon->ammo_use)
	{
	  /* search for a weapon this player can shoot with */
	  player_sendf (p, COL_LTGREY, "\nYou have run out of ammo for your %s.", weapon->name);
	  player_choose_weapon (p);
	  weapon = &(weapons[ p->weapon ]);
	  player_sendf (p, COL_LTGREY, "\nYou pull out your trusty %s.", weapon->name);
	}
    }
}

static const gchar *powerup_str[ NUM_POWERUPS ][ 2 ] =
{
  { "\n`%s's quad-damage fades", "\n`%s's quad-damage ran out" },
  { "\n`%s's protection fades", "\n`%s's protection ran out" },
  { NULL, "\n`%s' mysteriously appears in the room" },
  { "\n`%s's regeneration fades", "\n`%s's regeneration ran out" },
  { "\n`%s' begins to slow down", "\n`%s's haste ran out" },
  { NULL, "\n`%s's environment suit ran out" }
};

static void
_powerup_fade (Player_t *p, gint which)
{
  if (p->powerup[ which ] > 0)
    {
      p->powerup[ which ]--;
      if (p->powerup[ which ] == 0)
	{
	  if (powerup_str[ which ][ 1 ])
	    room_sendf (p->room, COL_EVENT, powerup_str[ which ][ 1 ], p->name);
	}
      else 
	if (p->powerup[ which ] < 5)
	  {
	    if (powerup_str[ which ][ 0 ])
	      room_sendf (p->room, COL_EVENT, powerup_str[ which ][ 0 ], p->name);
	  }
    }
}

static void
_player_think (Player_t *p, gpointer data)
{
  gint i;

  if (!p)
    return;

  server_ping ((Sock_t *) p);

  if (p->state != PLAYER_STATE_ALIVE)
    return;
  
  if (p->move < MOVE_MAX)
    {
      p->move = MAX(MOVE_MAX, p->move + ( p->powerup[ POWERUP_HASTE ] ? 2 : 1 ));
      if (p->move >= MOVE_COST && p->tell_me_when_i_can_move)
	{
	  player_send (p, COL_EVENT, "\nYou can move now.");
	  p->tell_me_when_i_can_move = FALSE;
	}
    }
  
  if (p->air < MAX_AIR)
    p->air++;

  if (p->say < SAY_MAX)
    {
      p->say++;
      if (p->say >= SAY_COST && p->tell_me_when_i_can_talk)
	{
	  player_send (p, COL_EVENT, "\nYou can talk/emote now.");
	  p->tell_me_when_i_can_talk = FALSE;
	}
    }

  if (p->powerup[ POWERUP_REGEN ] && p->health < 150)
    p->health = MIN( 150, p->health + 3 );
  
  for (i = 0; i < NUM_POWERUPS; i++)
    _powerup_fade (p, i);
}

void
players_think ()
{
  server_foreach ((SockFunc_t) _player_think, NULL);
}

void
player_send_prompt (Player_t *p)
{
  gint i;
  if (p->send_prompt == FALSE)
    return;
  if (p->state == PLAYER_STATE_ALIVE)
    {
      server_sendf ((Sock_t *) p, -1, "\nH:%d A:%d M:%d ", p->health, p->armor, p->move);
      if (weapons[ p->weapon ].ammo >= 0 && 
	  weapons[ p->weapon ].ammo_use)
      server_sendf ((Sock_t *) p, -1, "%s:%d ",
		    ammo_str[ weapons[ p->weapon ].ammo ],
		    p->ammo[ weapons[ p->weapon ].ammo ]);
      for (i = 0; i < NUM_POWERUPS; i++)
	if (p->powerup[ i ])
	  server_sendf ((Sock_t *) p, -1, "(%s) ", powerup_name[i]);
      if (p->peace_mode)
	server_send ((Sock_t *) p, COL_GREEN, "(peace-mode) ");
      server_send ((Sock_t *) p, -1, "> ");
    }
  else
    server_send ((Sock_t *) p, -1, "\n> ");
  p->send_prompt = FALSE;
}

/* */
static void
_player_send_prompt (Player_t *p, gpointer data)
{
  if (p->state != PLAYER_STATE_LOGIN)
    player_send_prompt (p);
}

void
players_send_prompts ()
{
  server_foreach ((SockFunc_t) _player_send_prompt, NULL);
}

void
player_send_stats (Player_t *p)
{
  gint i;
  time_t t = time (NULL);
  player_sendf (p, COL_LTGREY, "\n`%s'", p->name);
  for (i = 0; i < NUM_POWERUPS; i++)
    if (p->powerup[ i ])
      player_sendf (p, COL_LTGREY, " (%s)", powerup_name[ i ]);
  if (p->peace_mode)
    player_sendf (p, COL_GREEN, "\nIn peace mode");
  else if (p->attacking)
    player_sendf (p, -1, "\nAttacking: `%s'", p->attacking->name);
  player_sendf (p, -1, "\nFrags: %d Time: %u", p->frags, (t - p->join_time) / 60);
  player_sendf (p, -1, "\nHealth: %d Armor:%d Move:%d", p->health, p->armor, p->move);
  player_sendf (p, -1, "\nShells: %d Nails:%d Rockets:%d Slugs:%d Cells:%d",
		p->ammo[ AMMO_SHELLS ],
		p->ammo[ AMMO_NAILS ],
		p->ammo[ AMMO_ROCKETS ],
		p->ammo[ AMMO_SLUGS ],
		p->ammo[ AMMO_CELLS ]);

  player_sendf (p, COL_LTGREY, "\nWeapons:");
  for (i = 0; i < NUM_WEAPONS; i++)
    {
      if (player_has_weapon (p, i))
	{
	  player_sendf (p, -1, "\n (%d) %s", (i+1) % 10, weapons[i].name);
	  if (i == p->weapon)
	    player_send (p, -1, " (using)");
	}
    }
  if (p->keys)
    {
      player_sendf (p, COL_LTGREY, "\nKeys:");
      for (i = 0; i < NUM_KEYS; i++)
	{
	  if (player_has_key (p, i))
	    {
	      player_sendf (p, -1, "\n  %s skull-key", key_names[i]);
	    }
	}
    }
  player_send (p, -1, "\n");
}

void
player_give_ammo (Player_t *p, gint ammo[])
{
  gint i;
  for (i = 0; i < NUM_AMMO_TYPES; i++)
    p->ammo[ i ] = MIN(ammo_max_carry[i], p->ammo[i] + ammo[i]);
}

void
player_give_weapons (Player_t *p, gint _weapons)
{
  gint i, j;
  _weapons &= ~(p->weapons);
  if (_weapons)
    {
      for (i = WEAPON_BFG, j = 1 << i; 
	   i >= 0; 
	   i--, j >>=1)
	{
	  if (_weapons & j)
	    player_sendf (p, COL_EVENT, "\nYou got a %s.", weapons[i].name);
	}
      p->weapons |= _weapons;
    }
}

void
player_give_keys (Player_t *p, gint keys)
{
  gint i, j;
  keys &= ~(p->keys);
  if (keys)
    {
      for (i = 0, j = 1 << i; 
	   i < NUM_KEYS; 
	   i++, j <<=1)
	{
	  if (keys & j)
	    player_sendf (p, COL_EVENT, "\nYou got a %s skull-key.", key_names[i]);
	}
      p->keys |= keys;
    }
}

void
player_in_water (Player_t *p)
{
  if (p->powerup[ POWERUP_SUIT ]) return;
  if (p->air >= 2)
    p->air-=2;
  if ( p->powerup[ POWERUP_PENT ]) return;
  else 
    {
      player_send (p, COL_BAD, "\nGulp!  Gasp!  Gurgle!");
      if( 0 >= (p->health -= WATER_DAMAGE))
	{
	  game_broadcastf (COL_EVENT, "\n`%s' sleeps with the fishes.",  p->name);
	  player_died (p);
	}
    }
}

void
player_in_slime (Player_t *p)
{
  gint damage = SLIME_DAMAGE;
  if (p->powerup[ POWERUP_SUIT ])
    return;
  damage = _player_damage_armor (p, damage);
  if (p->powerup[ POWERUP_PENT ])
    return;
  player_send (p, COL_BAD, "\nAAAarrh, slime!");
  if (0 >= (p->health -= damage))
    {
      game_broadcastf (COL_EVENT, "\n`%s' gulped down some slime.", p->name);
      player_died (p);
    }
}

void
player_in_lava (Player_t *p)
{
  gint damage = LAVA_DAMAGE;
  if (p->powerup[ POWERUP_SUIT ])
    return;
  damage = _player_damage_armor (p, damage);
  if (p->powerup[ POWERUP_PENT ]) 
    return;
  player_send (p, COL_BAD, "\nAAAarrh, lava!");
  if (0 >= (p->health -= damage))
    {
      game_broadcastf (COL_EVENT, "\n`%s' did a back-flip in the lava.",  p->name);
      player_died (p);
    }
}

void
player_spawn (Player_t *p)
{
  Room_t *room;
  if (!p || p->state == PLAYER_STATE_ALIVE) return;
  if (!(room = room_spawn()))
    {
      player_sendf (p, COL_BAD, "No map running, or no rooms in map.");
      return;
    }
  player_set_defaults (p);
  room_sendf (room, COL_EVENT, "\n`%s' has spawned in the room.", p->name);
  player_sendf (p, COL_EVENT, "\nYou spawn in `%s'.\n", room->name);
  room_player_add (room, p);
  p->state = PLAYER_STATE_ALIVE;
  if (p->join_time == 0)
    p->join_time = time (NULL);
  room_look (room, p);
}


  
/* */
void
player_send (Player_t *p, gint col, gchar *text)
{
  server_send ((Sock_t *) p, col, text);
  p->send_prompt = TRUE;
}

void 
player_sendf (Player_t *p, gint col, gchar *format,...)
{
  gchar *text;
  va_list args;
  va_start (args, format);
  text = g_strdup_vprintf (format, args);
  va_end (args);
  server_sendf ((Sock_t *) p, col, text);
  p->send_prompt = TRUE;
  g_free (text);  
}

/* */
typedef struct _SendInfo _SendInfo_t;
struct _SendInfo {
  gint col;
  gchar * text;
};

static void
_player_send (Player_t * p, _SendInfo_t * data) {
  if (p->state != PLAYER_STATE_LOGIN)
    player_send (p, data->col, data->text);
}

void
players_send (gint col, gchar *text)
{
  _SendInfo_t info = { col, text };
  server_foreach ((SockFunc_t)_player_send, &info);
}

void
players_sendf (gint col, gchar * format,...)
{
  gchar *text;
  va_list args;
  va_start (args, format);
  text = g_strdup_vprintf (format, args);
  va_end (args);
  players_send (col, text);
  g_free (text);  
}

/* */
typedef struct _FindInfo _FindInfo_t;
struct _FindInfo {
  gchar * name;
  gint len;
  gboolean partial;
};

static gboolean
_player_find_with_name (Player_t *p, gpointer data)
{
  _FindInfo_t *_info = data;
  if (p->name == NULL)
    return FALSE;
  if (_info->partial)
    return  (0 == strncasecmp (p->name, _info->name, _info->len));
  return  (0 == strcasecmp (p->name, _info->name));
}

Player_t *
player_find (gchar *name, gboolean partial)
{
  _FindInfo_t _info;
  if (!name)
    return NULL;
  _info.name = name;
  _info.partial = partial;
  _info.len = strlen (name);
  return (Player_t *) server_find ((SockFindFunc_t) _player_find_with_name, &_info);
}
