/* QuakeMud, a silly mud using libhacknslash
 * Copyright (C) 2000 David Wynn (aka kismet), of the OSOAL collective
 * send abuse to:    david@osoal.org.nz
 * http requests to: web.osoal.org.nz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __PLAYERS_H__
#define __PLAYERS_H__

#include <glib.h>
#include <time.h>

#include <hacknslash/object.h>
#include <hacknslash/server.h>

typedef struct  Player              Player_t;
typedef struct  PlayerClass         PlayerClass_t;
typedef void (* PlayerInputFunc_t) (Player_t *p, gchar *line);

#define SAY_MAX    5
#define SAY_COST   1
#define SHOUT_COST 2

#define MOVE_MAX  9
#define MOVE_COST 4

#define WATER_DAMAGE 5
#define SLIME_DAMAGE 10
#define LAVA_DAMAGE  50

#define MAX_AIR 10

enum {
  PLAYER_STATE_LOGIN = 0, /* player is entering a character name */
  PLAYER_STATE_LIMBO,     /* player is ready to respawn */
  PLAYER_STATE_ALIVE,     /* player is running around */
  PLAYER_STATE_REMOVED
};

enum {
  EMOTE_WAVE = 0,
  EMOTE_FLIPOFF,
  EMOTE_SALUTE
};

enum {
  POWERUP_QUAD = 0,
  POWERUP_PENT,
  POWERUP_EYES,
  POWERUP_REGEN,
  POWERUP_HASTE,
  POWERUP_SUIT, /* environment suit, protects against water, slime and lava */

  NUM_POWERUPS
};

#include "rooms.h"
#include "weapons.h"
#include "items.h"

struct PlayerClass
{
  SockClass_t parent;
};

PlayerClass_t *player_class ();

extern const gchar *powerup_name[ NUM_POWERUPS ];

struct Player
{
  Sock_t parent;

  time_t join_time;
  gint state;

  /* input handler */
  PlayerInputFunc_t handler;

  gchar *name;
  gchar *team;
  gint   frags;

  /* flags */
  guint send_prompt             : 1;
  guint admin                   : 1;
  guint peace_mode              : 1;
  
  /* gotta love these long variables */ 
  guint tell_me_when_i_can_talk : 1;
  guint tell_me_when_i_can_move : 1;

  Room_t   *room;
  Player_t *attacking; /* the player this player is fighting */
  GSList   *attackers; /* players attacking this player */

  /* player above and below this player in the score ranking */
  Player_t *up, *down;

  /* say, move and air points */
  gint  say;
  gint  move;
  gint  air;

  gint  health;
  gint  armor;
  gint  weapon;  /* current weapon */
  guint weapons; /* mask of weapons this player has */
  gint  ammo[ NUM_AMMO_TYPES ];

  /* powerups */
  gint  powerup[ NUM_POWERUPS ];

  /* mask of keys this player has */
  guint keys;
};



/* */
void       players_think ();

void       player_set_name      (Player_t *p, gchar *name);
void       player_set_team      (Player_t *p, gchar *team);

void       player_spool         (Player_t *p, gchar *file);

Player_t * player_get           (guint     id);

void       player_to_game       (Player_t *p);
void       player_from_game     (Player_t *p);

/* spawn `p' in a random player-spawn room */
void       player_spawn         (Player_t *p);

/* returns the player with the most frags */
Player_t * player_winning();

void       player_set_defaults  (Player_t *p);

#define    player_has_weapon( PLAYER_PTR, WEAPON_NUM ) \
  ( (PLAYER_PTR)->weapons & ( 1 << WEAPON_NUM ) )

#define    player_has_key( PLAYER_PTR, KEY_NUM ) \
  ( (PLAYER_PTR)->keys & ( 1 << KEY_NUM ) )

void       player_give_weapons  (Player_t *p, gint weapons);

void       player_give_ammo     (Player_t *p, gint ammo[]);

void       player_give_keys     (Player_t *p, gint keys );

void       player_to_limbo      (Player_t *p);

/* send all players to limbo (for map changes) */
void       players_to_limbo     ();

Item_t *   player_backpack      (Player_t *p);

/* set `p' to attack `target' */
void       player_set_attack    (Player_t *p, Player_t *target );

void       player_attack        (Player_t *p, Player_t *target);

void       player_take_damage   (Player_t *p, Player_t *source, gint damage);

void       player_in_water      (Player_t *p);
void       player_in_slime      (Player_t *p);
void       player_in_lava       (Player_t *p);

void       player_suicide       (Player_t *p);

void       player_choose_weapon (Player_t *p);

void       player_from_room     (Player_t *p);
void       player_to_room       (Player_t *p, Room_t *room);

/* */
void       player_send_prompt   (Player_t *p);

/* sends `p' a list of all the connected players */
void       player_send_players  (Player_t *p);

/* send promps to all players that need them */
void       players_send_prompts ();

void       player_send_stats    (Player_t *p);

/* send high scores to `p' */
void player_send_scores (Player_t *p);

/* send text to all alive/limbo players */
void players_send  (gint col, gchar *text);
void players_sendf (gint col, gchar *format,...);

void player_send   (Player_t *p, gint col, gchar *text);
void player_sendf  (Player_t *p, gint col, gchar *format,...);

/* find the player called `name' */
Player_t *player_find (gchar *name, gboolean partial);

#endif /* __PLAYERS_H__ */
