/* QuakeMud, a silly mud using libhacknslash
 * Copyright (C) 2000 David Wynn (aka kismet), of the OSOAL collective
 * send abuse to:    david@osoal.org.nz
 * http requests to: web.osoal.org.nz
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>

#include <hacknslash/colors.h>
#include <hacknslash/token.h>
#include <hacknslash/misc.h>
#include <hacknslash/db.h>

#include "rooms.h"
#include "items.h"
#include "game.h"

/* */
const gchar *exit_dirs[ NUM_EXITS ] = 
{
  "north", "south", "east", "west", "up", "down", "teleport"
};

const gint   opp_dirs[ NUM_EXITS ] = 
{
  EXIT_SOUTH, EXIT_NORTH, EXIT_WEST, EXIT_EAST, EXIT_DOWN, EXIT_UP, EXIT_TELEPORT
};

const gchar *key_names[ NUM_KEYS ] = 
{
  "iron", "silver", "gold"
};

static gchar *room_flag[ NUM_ROOM_FLAGS ] = 
{
  "nokill",
  "water",
  "slime",
  "lava"
};



/* */
static ObjectClass_t *_exit_parent_class = NULL;

static void
_exit_init (Exit_t *exit)
{
  if (_exit_parent_class->init)
    _exit_parent_class->init ((Object_t *) exit);
  exit->name = NULL;
  exit->descript = NULL;
  exit->dir = -1;
  exit->key = -1;
  exit->rooms = NULL;
  exit->secret = FALSE;
}

static void
_exit_destroy (Exit_t *exit)
{
  g_free (exit->name);
  g_free (exit->descript);
  g_slist_free (exit->rooms);
  if (_exit_parent_class->destroy)
    _exit_parent_class->destroy ((Object_t *) exit);
}


static gboolean
_exit_add_var (Exit_t *exit, BlockVar_t *var)
{
  gint i;
  if (!strcmp (var->label, "name"))
    {
      g_free (exit->name);
      exit->name = g_strdup (var->value);
    }
  else if (!strcmp (var->label, "descript"))
    {
      g_free (exit->descript);
      exit->descript = g_strdup (var->value);
    }
  else if (!strcmp (var->label, "key"))
    {
      for (i = 0; i < NUM_KEYS; i++)
	{
	  if (!strcasecmp (var->value, key_names[i]))
	    {
	      exit->key = i;
	      break;
	    }
	}
    }
  else if (!strcmp (var->label, "dir"))
    {
      for (i = 0; i < NUM_EXITS; i++)
	{
	  if (!strcasecmp (var->value, exit_dirs[i]))
	    {
	      exit->dir = i;
	      break;
	    }
	}
    }
  else if (!strcmp (var->label, "room"))
    {
      exit->rooms = g_slist_prepend (exit->rooms, GINT_TO_POINTER(atoi (var->value)));
    }
  else if (!strcmp (var->label, "secret"))
    {
      misc_strtobool (&(exit->secret), var->value);
    }
  else
    return FALSE;
  
  return TRUE;
}

static ExitClass_t _exit_class;

ExitClass_t *
exit_class () {
  if (!_exit_parent_class)
    {
      ObjectClass_t *oclass = (ObjectClass_t *) &_exit_class;
      oclass->name = "Exit";
      oclass->size = sizeof (Exit_t);
      oclass->init = (ObjectFunc_t) _exit_init;
      oclass->destroy = (ObjectFunc_t) _exit_destroy;
      oclass->parent_class = _exit_parent_class = object_class ();
    }
  return &_exit_class;
}

/* */
static BlockDef_t _exit_block_def;

BlockDef_t *
exit_block_def () {
  static gboolean ran = FALSE;
  if (!ran)
    {
      ran = TRUE;
      _exit_block_def.oclass = (ObjectClass_t *) exit_class ();
      _exit_block_def.add_var = _exit_add_var;
      _exit_block_def.get_vars = NULL;
      _exit_block_def.blocks = NULL;
    }
  return &_exit_block_def;
}


/* */
static ObjectClass_t *_room_parent_class = NULL;

static void
_room_init (Room_t *room)
{
  if (_room_parent_class->init)
    _room_parent_class->init ((Object_t *) room);
  room->name = NULL;
  room->descript = NULL;
  room->players = NULL;
  room->items = NULL;
  room->player_spawn = FALSE;
  room->flags = 0;
}

static void
_room_destroy (Room_t *room)
{
  GSList *scan;
  g_free (room->name);
  g_free (room->descript);
  g_slist_free (room->players);
  for (scan = room->items; scan; scan = scan->next)
    item_free (scan->data);
  g_slist_free (room->items);
  if (_room_parent_class->destroy)
    _room_parent_class->destroy ((Object_t *) room);
}


/* */
static GSList *  _pspawn = NULL;
static gint      _pspawn_size = 0;
static DB_t      _room_db = { 100, NULL, NULL };
static Array_t * _rooms = NULL;


static gboolean
_room_add_var (Room_t *room, BlockVar_t *var)
{
  if (!strcmp (var->label, "name"))
    {
      g_free (room->name);
      room->name = g_strdup (var->value);
    }
  else if (!strcmp (var->label, "descript"))
    {
      g_free (room->descript);
      room->descript = g_strdup (var->value);
    }
  else if (!strcmp (var->label, "spawn"))
    {
      if (!strcmp (var->value, "player"))
	{
	  /* add room to the player spawn list */
	  if (room->player_spawn == FALSE)
	    {
	      room->player_spawn = TRUE;
	      _pspawn = g_slist_prepend (_pspawn, room);
	      _pspawn_size++;
	    }
	}
      else
	{
	  /* add item to room item list */
	  Item_t *item = item_new (var->value);
	  gint i;
	  if (item)
	    {
	      /* do not load powerup items if game_powerups flag is false */
	      if (!game_powerups)
		{
		  for (i = 0; i < NUM_POWERUPS; i++)
		    {
		      if (item->powerups[i])
			{
			  item_free (item);
			  item = NULL;
			  break;
			}
		    }
		}
	      if (item)
		{
		  item->spawn_time = 0;
		  room->items = g_slist_prepend (room->items, item);
		}
	    }
	} 
    }
  else
    {
      gint i;
      for (i = 0; i < NUM_ROOM_FLAGS; i++)
	{
	  if (!strcmp (var->label, room_flag[i]))
	    {
	      if (!strcasecmp (var->value, "true"))
		room->flags |= (1<<i);
	      else if (!strcasecmp (var->value, "false"))
		room->flags &= ((1<<(i+1))-1);
	      break;
	    }
	}
      /* else var is unrecognised, so return false */
      if (i == NUM_ROOM_FLAGS)
	return FALSE;
    }
  return TRUE;
}

/* */
static RoomClass_t _room_class;

RoomClass_t *
room_class ()
{
  if (_room_parent_class == NULL)
    {
      ObjectClass_t *oclass = (ObjectClass_t *) &_room_class;
      oclass->name = "Room";
      oclass->size = sizeof (Room_t);
      oclass->init = (ObjectFunc_t) _room_init;
      oclass->destroy = (ObjectFunc_t) _room_destroy;
      oclass->parent_class = _room_parent_class = object_class ();
    }
  return &_room_class;
}

/* */
static BlockDef_t _room_block_def;

BlockDef_t *
room_block_def ()
{
  static gboolean ran = FALSE;
  if (!ran)
    {
      ran = TRUE;
      _room_block_def.oclass = (ObjectClass_t *) room_class ();
      _room_block_def.add_var = _room_add_var;
      _room_block_def.get_vars = NULL; /* the room database is read-only */
      _room_block_def.blocks = g_slist_prepend (NULL, exit_block_def ());
    }
  return & _room_block_def;
}

static guint _room_event_types[ NUM_ROOM_EVENTS ];

guint
room_event_type (guint virtual_type)
{
  static gboolean ran = FALSE;
  if (!ran)
    {
      ran = TRUE;
      object_event_type_alloc (_room_event_types, NUM_ROOM_EVENTS);
    }
  return (virtual_type >= 0 && virtual_type < NUM_ROOM_EVENTS ?
	  _room_event_types[ virtual_type ] :
	  0 );
}



/* */
void
room_db_init ()
{
  static gboolean ran = FALSE;
  if (ran)
    return;
  ran = TRUE;
  _room_db.def = room_block_def ();
  _room_db.db_dir = NULL;
  _rooms = array_new (100, object_ref, object_unref);
}

void
room_db_load (gchar *dir)
{
  _room_db.db_dir = g_strdup (dir);
  db_load_all (&_room_db, _rooms);
}

void
room_db_free ()
{
  g_free (_room_db.db_dir);
  array_clear (_rooms);
  g_slist_free (_pspawn);
  _pspawn = NULL;
  _pspawn_size = 0;
}

Room_t *
room_db_get (guint id)
{
  return (Room_t *) array_get (_rooms, id);
}



/* */
void
room_send (Room_t *room, gint col, gchar *text)
{
  GSList *scan;
  if (!room)
    return;
  for (scan = room->players; scan; scan = scan->next)
    {
      Player_t *p = scan->data;
      player_send (p, col, text);
    }
}

void 
room_sendf (Room_t *room, gint col, gchar *format,...)
{
  gchar *text;
  va_list args;
  va_start (args, format);
  text = g_strdup_vprintf (format, args);
  va_end (args);
  room_send (room, col, text);
  g_free (text);
}

/* */
static void
_room_remove_item (Room_t *room, GSList **trail, GSList **scan)
{
  Item_t *item = (*scan)->data;
  GSList *next = (*scan)->next;
  if (*trail)
    (*trail)->next = next;
  else
    room->items = next;
  room->num_items--;
  item_free (item);
  g_slist_free_1 (*scan);
  *scan = next;
}

static void
_room_spawn_items (Room_t *room)
{
  GSList *scan, *trail;
  Item_t *item;
  time_t t = time (NULL);
  Player_t * p;

  if (room->items)
    {
      /* notify spawning of any items in the room */
      for (trail = NULL, scan = room->items; scan;)
	{
	  item = scan->data;
	  if (item->die_time > 0 && item->die_time < t)
	    {
	      /* unspawn items that have expired */
	      room_sendf (room, COL_EVENT, "\nA %s disapeared.", item->name);
	      if (item->spawn_interval)
		{
		  /* respawn item later */
		  item->spawn_time = t + item->spawn_interval;
		} 
	      else
		{
		  /* kill item as it won't respawn */
		  _room_remove_item (room, &trail, &scan);
		  continue;
		}
	    }
	  if (item->spawn_time < t && item->spawn_time > 0)
	    {
	      room_sendf (room, COL_EVENT, "\nA %s has just spawned.", item->name);
	      item->spawn_time = 0;
	    }
	  trail = scan; scan = scan->next;
	}
      
      /* give spawned items to random players */
      if (room->players)
	{
	  for (trail = NULL, scan = room->items; scan;)
	    {
	      item = scan->data;
	      if (item->spawn_time <= t)
		{
		  p = room_rand_player (room);
		  if (p && item_pickup (p, item))
		    {
		      if (item->spawn_interval)
			{
			  /* if item is set to respawn, reset spawn timer and leave
			   * item in list.
			   */
			  if (!game_weapon_persistance ||
			      item->always_pickup == TRUE ||
			      item->weapons == 0 )
			    item->spawn_time = t + item->spawn_interval;
			} 
		      else
			{
			  /* else item won't respawn, so remove from list and
			   * free it.
			   */
			  _room_remove_item (room, &trail, &scan);
			  continue;
			}
		    }
		}
	      trail = scan; scan = scan->next;
	    }
	}
    }
}

/* */
static void
_room_combat (Room_t *room)
{
  GSList *scan, *scan2, *trail;
  GSList *slist, *new;
  Player_t *p;

  /* do combat for this room */
  if (room->players && room->players->next)
    {
      
      /* this is too confusing */
      for (scan = room->players, slist = NULL;
	   scan;
	   scan = scan->next)
	{
	  p = scan->data;
	  
	  /* put player in a loosely random position in the list of attacking
	   * players.  all this is to stop frequent conditions where some
	   * players will almost always get to strike before others due to
	   * their id number or their position in the database.
	   *
	   * here the order of combat is governed by weapon speed.
	   * players with axes almost always strike last.
	   * players with very fast weapons like zapguns/railguns tend to strike
	   * first.
	   */
	  
	  if (p->peace_mode == FALSE)
	    {
	      new = g_slist_alloc ();
	      new->data = p;
	      /* make sure objects are alive for as long as we're going through this
	       * list.
	       */
	      object_ref ((Object_t *) p);
	      if (!slist)
		{
		  new->next = NULL;
		  slist = new;
		}
	      else
		{
		  for (scan2 = slist, trail = NULL;
		       ;
		       trail = scan2, scan2 = scan2->next)
		    {
		      /* this seems random enough in practice */
		      if (!scan2 || (random()%10) < (weapons[p->weapon].speed + (p->powerup[ POWERUP_HASTE ] ? 3 : 0 )) )
			{
			  if (trail)
			    trail->next = new;
			  else
			    slist = new;
			  new->next = scan2;
			  break;
			}
		    }
		}
	    }
	}
      
      if (slist)
	{
	  /* run through combat order */
	  for (scan = slist; scan; scan = scan->next)
	    {
	      Player_t * t;
	      p = scan->data;
	      if (p->attacking && p->attacking->room == room)
		t = p->attacking;
	      else
		t = room_rand_player_except (room, p->team);
	      player_attack (p, t);
	      object_unref ((Object_t *) p);
	    }
	  g_slist_free (slist);
	}
    }
}

/* */
static void
_room_do_effects_player (Room_t *room, Player_t *p)
{
  if (room->flags & (1<<ROOM_FLAG_WATER))
    player_in_water (p);
  else if (room->flags & (1<<ROOM_FLAG_SLIME))
    player_in_slime (p);
  else if (room->flags & (1<<ROOM_FLAG_LAVA))
    player_in_lava (p);
}

/* */
static void
_room_do_effects (Room_t *room)
{
  GSList *scan;
  for (scan = room->players; scan;)
    {
      GSList *next = scan->next;
      _room_do_effects_player (room, scan->data);
      scan = next;
    }
}

/* */
void
room_player_remove (Room_t *room, Player_t *player)
{
  if (!room || !player || player->room != room)
    return;
  room->num_players--;
  room->players = g_slist_remove (room->players, player);
  player->room = NULL;
}

void
room_player_add (Room_t *room, Player_t *player)
{
  if (!room || !player || player->room == room)
    return;
  room->num_players++;
  room->players = g_slist_prepend (room->players, player);
  player->room = room;
}

void
room_item_add (Room_t *room, Item_t *item)
{
  if (!room || !item)
    return;
  room->num_items++;
  room->items = g_slist_prepend (room->items, item);
}

/* */
void
room_think (Room_t *room, gint id, gpointer data)
{
  ObjectEvent_t event = {room_event_type (ROOM_EVENT_THINK)};

  if (!room)
    return;

  /* emit the think event */
  object_emit ((Object_t *) room, &event);

  _room_do_effects (room);
  if (!(room->flags & (1 << ROOM_FLAG_NOKILL)))
    _room_combat (room);
  _room_spawn_items (room);
}

void
rooms_think ()
{
  array_foreach (_rooms, (ArrayForFunc_t) room_think, NULL);
}

void
room_look (Room_t *room, Player_t *player)
{
  if (!room || !player)
    return;
  if (room->name)
    player_sendf (player, COL_LTGREY, "\n%s", room->name);
  else
    player_send (player, -1, "\nA very dark place");
  if (room->descript)
    player_sendf (player, -1, "\n%s", room->descript);
  room_exits_list (room, player);
  room_items_list (room, player);
  room_players_list (room, player);
  player_send (player, -1, "\n");
}

void
room_exits_list (Room_t *room, Player_t *player)
{
  GSList *scan;
  Object_t *o;
  gint i = 0;
  if (!room)
    return;
  for (scan = ((Object_t *) room)->objects; scan; scan = scan->next)
    {
      o = scan->data;
      /* scan through all child objects of `room' for exits */
      if (o->oclass == (ObjectClass_t *) exit_class ())
	{
	  Exit_t *exit = (Exit_t *) o;
	  Room_t *dest;
	  if (exit->secret || exit->rooms == NULL)
	    continue;
	  if (!i)
	    {
	      player_send (player, COL_LTGREY, "\nExits:");
	      i = 1;
	    }
	  if (exit->rooms->next == NULL)
	    {
	      dest = room_db_get (GPOINTER_TO_INT(exit->rooms->data));
	      if (dest)
		player_sendf (player, -1, "\n  `%s' -> %s", exit_dirs[ exit->dir ], dest->name);
	    }
	  else
	    {
	      player_sendf (player, -1, "\n  `%s' -> unknown", exit_dirs[ exit->dir ]);
	    }
	  if (exit->key >= 0)
	    player_send (player, -1, " (locked)");
	}
    }
  if (!i)
    player_send (player, -1, "\nThere are no exits.");
}

void
room_items_list (Room_t *room, Player_t *player)
{
  gint i;
  Item_t *item;
  GSList *scan;
  if (!room)
    return;
  if (room->items)
    {
      i = 0;
      for (scan = room->items; scan; scan = scan->next)
	{
	  item = scan->data;
	  if (item->spawn_time == 0)
	    {
	      if (!i)
		{
		  player_send (player, COL_LTGREY, "\nItems:");
		  i=1;
		}
	      player_sendf (player, -1, "\n  `%s'", item->name);
	    }
	}
    }
  if (!i)
    player_send (player, -1, "\nNo items.");
}

void
room_players_list (Room_t *room, Player_t *player)
{
  gint i, j;
  Player_t *p;
  GSList *scan;
  if (!room)
    return;
  if (room->players)
    {
      i = 0;
      for (scan = room->players; scan; scan = scan->next)
	{
	  p = scan->data;
	  if (p == player)
	    continue;
	  
	  if (!i)
	    {
	      player_send (player, COL_LTGREY, "\nPlayers:");
	      i = 1;
	    }
	  
	  if (p->powerup[ POWERUP_EYES ])
	    player_sendf (player, -1, "\n  a floating pair of eyes");
	  else
	    player_sendf (player, -1, "\n  `%s'", ((Player_t *) (scan->data))->name);
	  for (j = 0; j < NUM_POWERUPS; j++)
	    if (j != POWERUP_EYES)
	      if (p->powerup[ j ])
		player_sendf (player, COL_LTGREY, " (%s)", powerup_name[ j ]);
	}
    }
  if (!i)
    player_send (player, -1, "\nNo players.");
}

static Room_t *
_room_exit_dest (Room_t *room, gint dir, gint *key)
{
  GSList *scan;
  Object_t *o;
  if (!room)
    return NULL;
  for (scan = ((Object_t *) room)->objects; scan; scan = scan->next)
    {
      o = scan->data;
      if (o->oclass == (ObjectClass_t *) exit_class ()) 
	{
	  Exit_t *exit = (Exit_t *) o;
	  if (exit->dir == dir)
	    {
	      GSList *slist;
	      gint n;
	      *key = exit->key;
	      n = misc_slist_count (exit->rooms);
	      if (n)
		{
		  slist = g_slist_nth (exit->rooms, random () % n);
		  return room_db_get (GPOINTER_TO_INT(slist->data));
		}
	    }
	}
    }
  return NULL;
}

void
room_player_leave (Player_t *p, gint dir)
{
  RoomEventPlayerMove_t event = 
  { { room_event_type (ROOM_EVENT_PLAYER_MOVE) }, p, dir, FALSE };
  Room_t *room = p->room;

  room_player_remove (room, p);
  object_emit ((Object_t *) room, (ObjectEvent_t *) &event);

  if (p->powerup[ POWERUP_EYES ] == 0)
    {
      switch (dir)
	{
	case EXIT_TELEPORT:
	  room_sendf (room, COL_EVENT, "\n`%s' teleports from the room.", p->name);
	  break;
	  
	case EXIT_NORTH:
	case EXIT_SOUTH:
	case EXIT_EAST:
	case EXIT_WEST:
	  room_sendf (room, COL_EVENT,
		      "\n`%s' leaves the room via the %s.",
		      p->name,
		      exit_dirs[ dir ]);
	  break;
	  
	case EXIT_UP:
	  room_sendf( room, COL_EVENT, "\n`%s' ascends from the room.", p->name);
	  break;
	  
	case EXIT_DOWN:
	  room_sendf (room, COL_EVENT, "\n`%s' descends from the room.", p->name);
	  break;
	  
	default:
	  /* shouldn't really get here, but the default case is here anyway */
	  room_sendf (room, COL_EVENT, "\n`%s' mysteriously vanishes from the room.",  p->name);
	  break;
	}
    }
}

void
room_player_enter (Room_t *room, Player_t *p, gint dir)
{
  RoomEventPlayerMove_t event = 
  { { room_event_type (ROOM_EVENT_PLAYER_MOVE) }, p, dir, TRUE };
  
  if (p->powerup[ POWERUP_EYES ] == 0)
    {
      switch (dir)
	{
	case EXIT_TELEPORT:
	  room_sendf (room, COL_EVENT, "\n`%s' teleports into the room.", p->name);
	  break;
	  
	case EXIT_NORTH:
	case EXIT_SOUTH:
	case EXIT_EAST:
	case EXIT_WEST:
	  room_sendf (room, COL_EVENT,
		      "\n`%s' enters the room from the %s.",
		      p->name,
		      exit_dirs[ dir ]);
	  break;
	  
	case EXIT_UP:
	  room_sendf (room, COL_EVENT, "\n`%s' enters the room from below.", p->name);
	  break;

	case EXIT_DOWN:
	  room_sendf (room, COL_EVENT, "\n`%s' enters the room from above.", p->name);
	  break;
	  
	default:
	  /* shouldn't really get here, but the default case is here anyway */
	  room_sendf (room, COL_EVENT, 
		      "\n`%s' somehow appears in the middle of the room.",
		      p->name);
	  break;
	}
    }
  
  object_emit ((Object_t *) room, (ObjectEvent_t *) &event);
  room_player_add (room, p);
  room_look (room, p);
  _room_do_effects_player (room, p);
}

void
room_player_move (Player_t *p, gint dir)
{
  Room_t *room, *old;
  gint key;

  if (p->room == NULL)
    return;
  old = p->room;

  room = _room_exit_dest (p->room, dir, &key);
  if (!room)
    {
      if (dir == EXIT_TELEPORT)
	player_send (p, COL_BAD, "There is no teleport in this room.");
      else
	player_send (p, COL_BAD, "You cannot go in that direction.");
      return;
    }
  if (p->move < MOVE_COST)
    {
      player_send (p, COL_BAD, "You are to tired to move right now.");
      p->tell_me_when_i_can_move = TRUE;
      return;
    }
  if (key >= 0 && !player_has_key (p, key))
    {
      player_sendf (p, COL_BAD,
		    "You need the %s skull-key to go through that door.",
		    key_names[key] );
      return;
    }

  p->move -= MOVE_COST;

  room_player_leave (p, dir);
  room_player_enter (room, p, opp_dirs[ dir ]);
}

Room_t *
room_spawn ()
{
  GSList *slist;
  if (!_pspawn_size)
    return NULL;
  slist = g_slist_nth (_pspawn, random () % _pspawn_size);
  if (!slist)
    return NULL;
  return (Room_t *) (slist->data);
}

Player_t *
room_player_find (Room_t *room, gchar *name)
{
  GSList *scan;
  gint best_len = 0, i;
  Player_t *best = NULL;
  if (!room)
    return NULL;
  for (scan = room->players; scan; scan = scan->next)
    {
      Player_t *p = scan->data;
      if (p->powerup[ POWERUP_EYES ]) 
	continue;
      for (i = 0; p->name[i] == name[i] && name[i]; i++);
      if (!(name[i]))
	return p;
      if (i > best_len)
	{
	  best_len = i;
	  best = p;
	}
    }
  return best;
}

Player_t *
room_rand_player (Room_t *room)
{
  gint n;
  GSList *slist;
  if (!room)
    return NULL;
  n = room->num_players;
  if (n)
    {
      slist = g_slist_nth (room->players, random () % n);
      return (Player_t *) (slist->data);
    }
  return NULL;
}

static gint
_room_players_count (Room_t *room, gchar *team)
{
  gint count = 0;
  GSList *scan;
  for (scan = room->players; scan; scan = scan->next)
    {
      Player_t * p = scan->data;
      if (!strcasecmp (p->team, team))
	count++;
    }
  return count;
}

Player_t *
room_rand_player_except (Room_t *room, gchar *team)
{
  gint n;
  GSList *slist;
  Player_t *p;
  if (!room)
    return NULL;
  n = room->num_players - _room_players_count (room, team);
  if (n < 1)
    return NULL;
  n = random () % n;
  for (slist = room->players; slist; slist = slist->next)
    {
      p = slist->data;
      if (!strcasecmp (p->team, team))
	continue;
      if (!n)
	return p;
      n--;
    }
  return NULL;
}
